/* this is the header for ar.c that works with the ms omf that watcom
   uses.  tested under QNX 4. -- mph */

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>

// support for omf archives

static int page_len ;

struct omf_rec
{
	unsigned char type ;
	unsigned short len ;
	char data[] ;
	// followed by checksum
} ;

struct objinfo
{
	char *name ; 	// file name of object (malloced)
	int date ;	// date on file
	long position ;	// position of object in library
} ;

struct libinfo
{
    char *lib_name;	// name of lib (malloced)
    struct objinfo *d;	// array of entries in the lib (malloced)
    int valid;		// number of valid entries
    int allocated;	// number of entries allocated
    struct stat stat;	// stat of lib
};

struct omf_theadr
{
	unsigned char string_length ;
	char string[] ;
	// char checksum ;
} ;

struct omf_comment
{
	unsigned char comment_type:6 ; // should be 0
	unsigned char np:1 ;
	unsigned char nl:1 ;
	unsigned char comment_class ;
	unsigned char commentary[] ;
	// followed by checksum
} ;

struct omf_watcom_timestamp
{ 	// this structure lies over the commentary member of omf_comment when
 	// commet_class == 0xfe
	unsigned char timestamp_flag ; // should be 'T'
	unsigned int timestamp_data ;
	unsigned char checksum ;
} ;

struct omf_rec *get_omf_rec(FILE *fp, struct omf_rec *rec, int rec_size)
{
	static struct omf_rec *lrec= 0 ;
	static unsigned short llen= 0 ;
	struct omf_rec *r ;	// record we are manipulating
	int s ; 		// allocated size of the record we are manip
	int amt ;

	if ((rec == 0) || (rec_size <= 3))
	{ // if we are useing the local storage then make sure it is set up
		if (lrec == 0)
		{
			int size= (rec_size <= 1024) ? 1024 : rec_size ;
			lrec= malloc(size) ;
			if (lrec == 0) return 0 ;
			llen= size ;
		} else
		{
			if (llen < rec_size)
			{
				struct omf_rec *tmp ;
				if ((tmp= realloc(lrec, rec_size)) == 0)
					return 0 ;
				lrec= tmp ;
				llen= rec_size ;
			}
		}
		r= lrec ;
		s= llen ;
	} else
	{ // we were handed storage
		r= rec ;
		s= rec_size ;
	}

	r->type= fgetc(fp) ;
	fread(&r->len, sizeof(r->len), 1, fp) ;

	// save some info if we are manipulating a library
	if (r->type == 0xF0)
		page_len= r->len+3 ;

	amt= r->len <= (s - 3) ? r->len : s - 3 ;
	fread(r->data, amt, 1, fp) ;

	// for omf_arscan.c this is only meant to manipulate libs
	if (((r->type == 0x8A) || (r->type == 0x8B)))
	{
		int n= page_len - ftell(fp)%page_len ;
		if (n != page_len)
		    while(n-- > 0) 
			fgetc(fp) ;
	} else
		while(amt++ < r->len) fgetc(fp) ;


	return r ;
}

const struct objinfo *nextobj(FILE *fp)
{
	static struct objinfo rv ; /* return value */
	struct omf_rec *r ;
	struct omf_theadr *theadr ;

	if (fp == 0) return 0 ;

	do
	{
	    rv.position= ftell(fp) ;
	    r= get_omf_rec(fp, 0, 0);
	} while(r != 0 && r->type != 0x80 && r->type != 0xf1);
	if (r == 0 || r->type == 0xf1) 
	    return 0;

	theadr= (struct omf_theadr *)r->data ;
	rv.name= malloc(sizeof(char)*(theadr->string_length+1)) ;
	if (rv.name == 0) return 0 ;
	memcpy(rv.name, theadr->string, theadr->string_length) ;
	rv.name[theadr->string_length]= 0 ;

	rv.date= 0 ; // zero out date in case it doesn't exist
	do
	{
		r= get_omf_rec(fp, 0, 0) ;
		if (r == 0) break ;
		if (	(r->type == 0x88) &&
			(((struct omf_comment *)r->data)->np) &&
			(((struct omf_comment *)r->data)->nl))
		{
			struct omf_comment *c= (struct omf_comment *)r->data ;
			rv.date= ((struct omf_watcom_timestamp *)c->commentary)->timestamp_data ;
		}
	} while((r->type != 0x8a) && (r->type != 0x8b)) ; // stop at modend

	return (const struct objinfo *)&rv ;
}

static void free_lib(struct libinfo *li)
{
    if (li->lib_name)
    {
	free(li->lib_name);
	li->lib_name= 0;
    }
    if (li->d)
    {
	free(li->d);
	li->d= 0;
    }
    li->valid= li->allocated= 0;
}

static int slurp_lib(FILE *fp, struct libinfo *li)
{
    struct objinfo *o;

    assert(li->lib_name == 0);
    assert(li->valid == 0);

    while((o= nextobj(fp)) != 0)
    {
	char *tmp;
	int d;
	tmp= strrchr(o->name, '.') ;
	if (tmp == 0)
	{
	    void *t;
	    t= realloc(o->name, strlen(o->name)+2);
	    if (t)
		o->name= t;
	    else
	    {
		free(o->name);
		o->name= 0;
		free_lib(li);
		return -1;
	    }
	    tmp= o->name + strlen(o->name);
	    *tmp= '.';
	}
	*(tmp+1)= 'o' ; // omf libs members have .c extentions (sometimes?)
	*(tmp+2)= 0;
#if 0
	tmp= strrchr(o->name, '//') ;
	if (tmp == 0)
		tmp= o->name ;
	else
		tmp++ ; // one past path separator
	o->name= tmp;
#endif

	if (li->valid >= li->allocated)
	{
	    int sz= li->allocated ? 2*li->allocated : 64;
	    void *tmp= realloc(li->d, sz*sizeof(*li->d));
	    if (tmp)
	    {
		li->d= tmp;
		li->allocated= sz;
	    }
	    else
	    {
		free_lib(li);
		return 1;
	    }
	}
	li->d[li->valid++]= *o;
    }
    return 0;
}

int touchobj(FILE *fp, struct objinfo *o)
{
	struct omf_rec *r ;
	int found= 0 ;
	int pos= 0 ;

	if (fp == 0) return 0 ;

	if (fseek(fp, o->position, SEEK_SET) == -1)
		return 0 ;

	r= get_omf_rec(fp, 0, 0) ;
	if ((r == 0) || (r->type != 0x80)) return 0 ;

	do
	{
		pos= ftell(fp) ;
		r= get_omf_rec(fp, 0, 0) ;
		if (r == 0) break ;
		if (	(r->type == 0x88) &&
			(((struct omf_comment *)r->data)->np) &&
			(((struct omf_comment *)r->data)->nl))
		{
			int t= ftell(fp) ;
			if (t != -1)
			{ 
				found= 1 ;
				if (fseek(fp, pos+6, SEEK_SET) != -1)
				{
					int now= time(0) ;
					fwrite(&now, sizeof(now), 1, fp) ;
					fseek(fp, t, SEEK_SET) ;
				}
			}
		}
	} while((r->type != 0x8a) && (r->type != 0x8b)) ; // stop at modend

	if (!found) // we didn't find and modify a time comment record
	{ // force the date to change to now
		int t= ftell(fp) ;
		int c ;
		if (t == -1) return 0 ;
		c= fgetc(fp) ;
		if (fseek(fp, t, SEEK_SET) == -1) return 0 ;
		fputc(c, fp) ;
		if (fseek(fp, t, SEEK_SET) == -1) return 0 ;
	}

	return 1 ;
}


/* Takes three arguments ARCHIVE, FUNCTION and ARG.

   Open the archive named ARCHIVE, find its members one by one,
   and for each one call FUNCTION with the following arguments:
     archive file descriptor for reading the data,
     member name,
     member name might be truncated flag,
     member header position in file,
     member data position in file,
     member data size,
     member date,
     member uid,
     member gid,
     member protection mode,
     ARG.

   The descriptor is poised to read the data of the member
   when FUNCTION is called.  It does not matter how much
   data FUNCTION reads.

   If FUNCTION returns nonzero, we immediately return
   what FUNCTION returned.

   Returns -1 if archive does not exist,
   Returns -2 if archive has invalid format.
   Returns 0 if have scanned successfully.  */

long int
ar_scan (archive, function, arg)
     char *archive;
     long int (*function) ();
     long int arg;
{
    	FILE *fp= fopen(archive, "r") ;
    	int member_offset, data_offset ;
    	int length, page, type, i;
    	char junk ;
    	char *tmp ;
    	struct stat stat_buf ;
    	int fnval ;
	struct omf_rec *r ;
	static struct libinfo libcache= {0, 0, 0, 0};

    	if (fp == 0) return -1 ;
	if (fstat(fileno(fp), &stat_buf) == -1) 
	{
	    fclose(fp);
	    return -1 ;
	}

	r= get_omf_rec(fp, 0, 0) ;
	if (r->type != 0xf0)
	{ // not a library
		fclose(fp) ;
		return -2 ;
	}

	if (!libcache.lib_name || libcache.stat.st_mtime != stat_buf.st_mtime)
	{
fprintf(stderr, "reading lib: %s %d %d\n", libcache.lib_name, libcache.stat.st_mtime, stat_buf.st_mtime);
	    free_lib(&libcache);
	    if (!slurp_lib(fp, &libcache))
	    {
		libcache.lib_name= strdup(archive);
		libcache.stat= stat_buf;
	    }
	}

	if (!libcache.lib_name)
	    return -2;

	for(i= 0;i < libcache.valid; i++)
	{
		char *tmp ;
		int d ;
#if 0
		tmp= strrchr(libcache.d[i].name, '//') ;
		if (tmp == 0)
			tmp= libcache.d[i].name;
		else
			tmp++ ; // one past path separator
#else
		tmp= libcache.d[i].name;
#endif
		d= (libcache.d[i].date == 0) ? stat_buf.st_mtime : libcache.d[i].date ;
//fprintf(stderr, "omf_arscan: calling function on %s at %d %s\n",
//	tmp, d, ctime(&d));
		fnval= function(fileno(fp), tmp, 0,
				libcache.d[i].position, 
				libcache.d[i].position, 
				0, d,
				stat_buf.st_uid, stat_buf.st_gid,
				0, arg) ;
		if (fnval)
		{
	    		fclose(fp) ;
	    		return fnval ;
		}
    	}
	fclose(fp);
 	return 0 ;
}


/* Return nonzero iff NAME matches MEM.
   If TRUNCATED is nonzero, MEM may be truncated to
   sizeof (struct ar_hdr.ar_name) - 1.  */

int
ar_name_equal (name, mem, truncated)
     char *name, *mem;
     int truncated;
{
  return !strcmp (name, mem);
}

/* ARGSUSED */
static long int
ar_member_pos (desc, mem, truncated,
	       hdrpos, datapos, size, date, uid, gid, mode, name)
     int desc;
     char *mem;
     int truncated;
     long int hdrpos, datapos, size, date;
     int uid, gid, mode;
     char *name;
{
  if (!ar_name_equal (name, mem, truncated))
    return 0;
  return hdrpos;
}

/* Set date of member MEMNAME in archive ARNAME to current time.
   Returns 0 if successful,
   -1 if file ARNAME does not exist,
   -2 if not a valid archive,
   -3 if other random system call error (including file read-only),
   1 if valid but member MEMNAME does not exist.  */

int
ar_member_touch (arname, memname)
     char *arname, *memname;
{
	FILE *fp= fopen(arname, "r+b") ; // binary for compat with messy dog
	struct objinfo *o ;

	if (fp == 0) return -1 ;

	while((o= nextobj(fp)) != 0)
	{
		char *lc ;
		lc= strrchr(o->name, '//') ;
		if (lc == 0)
			lc= o->name ;
		else
			lc++ ;
		if (strcmp(lc, memname) == 0)
		{
			touchobj(fp, o) ;
			free(o->name) ;
			fclose(fp) ;
			return 0 ;
		}
		free(o->name) ;
	}
	return 1 ;
}

long int
describe_member (desc, name, truncated,
		 hdrpos, datapos, size, date, uid, gid, mode)
     int desc;
     char *name;
     int truncated;
     long int hdrpos, datapos, size, date;
     int uid, gid, mode;
{
  extern char *ctime ();

  printf ("Member `%s'%s: %ld bytes at %ld (%ld).\n",
	  name, truncated ? " (name might be truncated)" : "",
	  size, hdrpos, datapos);
  printf ("  Date %s", ctime (&date));
  printf ("  uid = %d, gid = %d, mode = 0%o.\n", uid, gid, mode);

  return 0;
}

main (argc, argv)
     int argc;
     char **argv;
{
  ar_scan (argv[1], describe_member, 0);
  return 0;
}


